-- exw_parser.e
--
-- reads the input file using ox generated parser
-- and reports interesting things.
-- ide_prj.e will then take the results and try to generate
-- a .prj file.

include ox_ide_p.e -- the parser generated by Ox

object result

procedure add_event( sequence s)
	registered_events &= {s}
end procedure

procedure add_control( sequence s)
	created_controls &= {s}
end procedure

procedure add_routine( sequence s)
	routines &= {s & {{}, 0} } -- the end will be added later by set_routine_end()
end procedure

procedure set_routine_end( sequence s)
	routines[length(routines)][ROUTINE_END] = s
end procedure

procedure add_ignored( sequence s)
	ignored &= {s}
end procedure

function try_RegisterEvent()
	result = oxParse( "RegisterEvent" )
	if length(result) and result[1] = RegisterEvent then
		if DEBUG then
			printf(log_fn, "RegisterEvent: %s %s %s\n", result[6..length(result)])
		end if
		add_event(result[2..length(result)])
		add_ignored(result[2..5])
		return 1
	end if
	return 0
end function

function try_ControlCreation()
	result = oxParse( "ControlCreation" )
	if length(result) = 6 and sequence (result[6]) and result[1] = ControlCreation then
		add_ignored(result[2..5])
		for i=1 to length(result[6]) do
			--printf(log_fn, "ControlCreation: %s %d %s %s %s %d %d %d %d\n",
			--	result[6][i][1..3] & "<multi>" & result[6][i][5..length(result[6][i])] )
			add_control(result[6][i])
		end for
		return 1
	end if
	return 0
end function

function try_Proc()
	result = oxParse( "Proc" )
	if length(result) and result[1] = Proc then
		if DEBUG then
			printf(log_fn, "Proc: %d %d %d %d %s\n", result[2..6])
		end if
		add_routine({result[6], result[2..5]})
		-- we're addinng it to ingored later
		return 1
	end if
	return 0
end function

function try_EndProc()
	result = oxParse( "EndProc" )
	if length(result) and result[1] = EndProc then
		if DEBUG then
			printf(log_fn, "EndProc: %d %d %d %d\n", result[2..5])
		end if
		set_routine_end(result[2..5])
		return 1
	end if
	return 0
end function

function try_Junk()
	result = oxParse( "Junk" )
	if length(result) and (result[1] = Junk or result[1] = WinMain) then
		if DEBUG then
			printf(log_fn, "Junk: %d %d %d %d\n", result[2..5])
		end if
		add_ignored(result[2..5])
		if result[1] = WinMain then
			if length(winmain) then
				parseWarning("File has more than one WinMain(). Only the last will be used.\n")
			end if
			winmain = result[6..7]
		end if
		return 1
	end if
	return 0
end function

sequence parse_routines
parse_routines = {}
parse_routines &= routine_id("try_RegisterEvent")
parse_routines &= routine_id("try_ControlCreation")
parse_routines &= routine_id("try_Proc")
parse_routines &= routine_id("try_EndProc")
parse_routines &= routine_id("try_Junk")

function try_tokens()
	-- i have to save the ox state and then pop it again
	-- after each try, because when Ox thinks token is matching
	-- and then encounter a bug, it won't get to the start of the
	-- token => some tokens would get missed.	
	
	for i=1 to length(parse_routines) do
		oxMySaveState()
		if call_func(parse_routines[i], {}) then
			return 0
		end if
		oxUnloadBuffer()
	end for

	-- this token is not interesting, move onto the next
	oxMoveAhead()
	if oxTokenIs("<EOF>") then
		-- end of file, stop parsing
		return 1
	end if
	return 0
end function

global procedure parse_file()
	while 1 do
		if try_tokens() then
			exit
		end if
	end while
end procedure
