-- ide_prj.e
--
-- writes the output .prj file for IDE

include sort.e  -- custom_sort()
include order.e -- for ordering controls
include file.e

constant
	-- for routines
	START_CHAR = 1,
	START_LINE = 2,
	END_CHAR   = 3,
	END_LINE   = 4

procedure convertWarning(sequence s)
	printf(1, "Conversion warning: %s", {s})
end procedure

procedure write(object s)
	puts(out_fn, s)
	if DEBUG then
		flush(out_fn)
	end if
end procedure

procedure writef(sequence format, object data)
	printf(out_fn, format, data)
	if DEBUG then
		flush(out_fn)
	end if
end procedure
	
function escape_string(sequence s)
	sequence res
	integer si
	res = ""
	for i=1 to length(s) do
		si = s[i]
		if    si= '\t' then res &= "\\t"
		elsif si= '\n' then res &= "\\n"
		elsif si= '\r' then res &= "\\r"
		elsif si= '\\' then res &= "\\\\"
		-- IDE doesn't need " to be escaped
		-- elsif si= '\"' then res &= "\\""
		else
			res &= si
		end if
	end for
	return res
end function

-- to properly sort ignored
function compare_ignored(sequence s1, sequence s2)
	if s1[END_LINE] < s2[END_LINE] then
		return -1
	elsif s1[END_LINE] > s2[END_LINE] then
		return  1
	else
		if s1[END_CHAR] < s2[END_CHAR] then
			return -1
		else
			return  1
		end if
	end if
end function
integer compare_ignored_id
compare_ignored_id = routine_id("compare_ignored")

procedure write_chunk(integer start_char,integer start_line,
	integer end_char, integer end_line )

	sequence out_line
	
	for line = start_line to end_line do
		if line = start_line then
			out_line = input_file_lines[line][start_char..length(input_file_lines[line])]
		elsif line = end_line then
			out_line = input_file_lines[line][1..end_char]
		else
			out_line = input_file_lines[line]
		end if
		write(out_line)
	end for
	
end procedure

----------------------------------------------------------
-- writes event (id points to item in registered_events)
----------------------------------------------------------
integer dummy_count
dummy_count = 0

procedure write_event(integer id)
	integer start_line, end_line,
	start_char, end_char, this_routine_id
	sequence this_routine, this_event

	this_event = registered_events[id]
	-- find corresponding routine for this event
	for i=1 to length(routines) do
		if equal(this_event[EVENT_ROUTINE], routines[i][ROUTINE_NAME]) then
			this_routine = routines[i]
			this_routine_id = i
			exit
		end if
	end for

	start_line = this_routine[ROUTINE_START][START_LINE]
	end_line   = this_routine[ROUTINE_END][END_LINE]
	start_char = this_routine[ROUTINE_START][START_CHAR]
	end_char   = this_routine[ROUTINE_END][END_CHAR]
	
	writef("#Proc: %s"&LF, {this_event[EVENT_NAME]})
	
	if this_routine[ROUTINE_USED] then 
		-- already written, two controls share the same routine
		writef ("procedure dummy_%d ()"&LF, dummy_count)
		write ("\t--this is needed to fool IDE"&LF)
		write ("end procedure"&LF)
		write("--exw2prj: handler for this event is already written"&LF)
		dummy_count +=1
	else	
		write_chunk(start_char,start_line, end_char, end_line )
		write(LF)
		routines[this_routine_id][ROUTINE_USED] = 1 -- mark as used
	end if
	
	write_chunk( this_event[START_CHAR], this_event[START_LINE],
	this_event[END_CHAR],this_event[END_LINE])
	
	write( LF&"#EndProc"&LF&LF)
end procedure

----------------------------------------------------------
-- for given name, returns index of created_controls
----------------------------------------------------------
function control_by_name(sequence name)
	for i=1 to length(created_controls) do
		if equal(created_controls[i][CONTROL_NAME], name) then
			return i
		end if
	end for
	convertWarning(sprintf("Error: Control %s was not found.", {name}))
	-- runtime ERROR here
end function

----------------------------------------------------------
-- find the highest parent of a control
----------------------------------------------------------
function find_top_parent_id(integer id)
	integer parent

	parent = created_controls[id][CONTROL_PARENTID]
	if parent = 0 or equal(created_controls[id][CONTROL_CLASS], "Window") then
		return id
	else
		return find_top_parent_id(parent)
	end if
end function

----------------------------------------------------------
-- for a control that is in TabItem, ide needs the TabItem
-- as parent, even if the control isn't direct child of the TabItem
----------------------------------------------------------
function Find_tabitem_parent(integer id)
	integer parent

	parent = created_controls[id][CONTROL_PARENTID]
	if parent = 0 then
		return id
	elsif equal(created_controls[id][CONTROL_CLASS], "TabItem")then
		return id
	else
		return Find_tabitem_parent(parent)
	end if
end function

function find_tabitem_parent(integer id)
	integer ok
	integer parent
	ok = Find_tabitem_parent(id)
	if not equal(created_controls[ok][CONTROL_CLASS], "TabItem" ) then
		parent = created_controls[id][CONTROL_PARENTID]
		return parent
	else
		return ok
	end if
end function

----------------------------------------------------------
-- ide needs absolute coords in a window, while create() uses
-- relative to control's parent
----------------------------------------------------------
function get_absolute(integer dimension, integer id)
	integer parent

	parent = created_controls[id][CONTROL_PARENTID]
	if parent = 0 then
		return created_controls[id][dimension]
	else
		if equal( created_controls[parent][CONTROL_CLASS] , "Window") then
			-- the parent is a Window, don't add
			return created_controls[id][dimension]
		end if
		return created_controls[id][dimension] + get_absolute(dimension, parent)
	end if
end function

constant
	Popup     = 1,
	PopupItem = 2,
	Menu      = 3,
	MenuItem  = 4

-- another nice routine :-)
-- for given control returns {menu type, subpopup/submenu depth}
-- or 0 if not menu/menuitem
function popup_info(integer id, integer level, integer class)
	integer parent
	sequence seq_class
	parent    = created_controls[id][CONTROL_PARENTID]
	seq_class = created_controls[id][CONTROL_CLASS]
	if equal(seq_class, "MenuItem") then
		if class = 0 then class = PopupItem end if
	elsif equal(seq_class, "Menu") then
		if equal(created_controls[parent][CONTROL_CLASS], "Window") then
			-- a menu
			if class = 0 then class = Popup end if
			level += 1
			if class = Popup then
				class = Menu
			elsif class = PopupItem then
				class = MenuItem
			end if
			return {class, level}
		else
			if class = 0 then class = Popup end if
			level += 1
		end if
	elsif equal(seq_class, "Popup") then
		if class = 0 then class = Popup end if
		level += 1
		return {class, level}
	else
		return 0
	end if
	return popup_info(parent, level, class)
end function

----------------------------------------------------------
-- writes a control to the file
-- "Judith ! The prj files are so weird !" :-)
----------------------------------------------------------
sequence tab_controls 
	-- holds info about tabcontrols id, and # of childs
tab_controls = {}

procedure write_control(integer id)
	object ok
	sequence this
	integer tab_parent
	atom x,y
	integer xy_default
	object popupinfo
	
	this = created_controls[id]

	-- if it is a TabControl, add it to our tab list
	if equal(this[CONTROL_CLASS], "TabControl") then
		tab_controls &= {{this[CONTROL_NAME], 1, id}}
		tab_parent = length(tab_controls)
	end if

	if equal(this[CONTROL_CLASS], "TabItem") then
		tab_parent = 0
		for i=1 to length(tab_controls) do
			if equal(this[CONTROL_PARENT], tab_controls[i][1]) then
				tab_parent = i	
				exit
			end if
		end for
		tab_controls[tab_parent][2] += 1
	end if

	-- the attributes

	write( "#Control : " & this[CONTROL_NAME] &LF)

	if this[CONTROL_CAPTION][CAPTION_STYLE] = CAPTION_SINGLE then
		write( "Title   : "  & escape_string(this[CONTROL_CAPTION][CAPTION_VALUE]) &LF)
	else
		-- ListViews multiple captions
		write( "Title   : {")
		ok = length(this[CONTROL_CAPTION][CAPTION_VALUE])
		for i=1 to ok-1 do
			write( '"' & escape_string(this[CONTROL_CAPTION][CAPTION_VALUE][i]) & "\"," )
		end for
		write( '"' & escape_string(this[CONTROL_CAPTION][CAPTION_VALUE][ok]) & "\"}")
		write(LF)
	end if

	popupinfo = popup_info(id, 0, 0)
	-- IDE has special names for popups different from win32lib
	-- and special names for submenus and subpopups
	if sequence(popupinfo) then
		if popupinfo[2] = 1 then
			if popupinfo[1] = Popup then
				write(   "Class   : Popup" &LF)
			elsif popupinfo[1] = PopupItem then
				write(   "Class   : PopupItem" &LF)
			elsif popupinfo[1] = Menu then
				write(   "Class   : Menu" &LF)
			elsif popupinfo[1] = MenuItem then
				write(   "Class   : MenuItem" &LF)
			end if
		else
			if popupinfo[2] = 2 then
				ok = ""
			else
				if popupinfo[2] > 4 then
					convertWarning("IDE doesn't support subpopupus/submenus"
					& "with level higher than 4")
				end if	
				ok = sprintf("%d", popupinfo[2]-1)				
			end if
			if    popupinfo[1]  = Popup then
				write(   "Class   : Sub" & ok & "popup" &LF)
			elsif popupinfo[1] = PopupItem then
				write(   "Class   : Sub" & ok & "popupItem" &LF)
			elsif popupinfo[1] = Menu then
				write(   "Class   : Sub" & ok & "menu" &LF)
			elsif popupinfo[1] = MenuItem then
				write(   "Class   : Sub" & ok & "menuItem" &LF)
			end if
		end if	
	else		
		write(   "Class   : "  & this[CONTROL_CLASS] &LF)
	end if
	
	if this[CONTROL_X] = Default or this[CONTROL_Y] = Default then
		x = 0
		y = 0
		xy_default = 1
	else
		x = get_absolute(CONTROL_X, id)
		y = get_absolute(CONTROL_Y, id)
		xy_default = 0
	end if
	
	if equal(this[CONTROL_CLASS], "TabItem" ) then
		writef("X       : %g"&LF, {x} )
		writef("Y       : %g"&LF, {y} )
		-- although TabItems have no CX or CY, IDE needs CX&CY of parent TabControl here.
		writef("CX      : %g"&LF, {created_controls[tab_controls[tab_parent][3]][CONTROL_CX]} )
		writef("CY      : %g"&LF, {created_controls[tab_controls[tab_parent][3]][CONTROL_CY]} )
	elsif find(this[CONTROL_CLASS], {"Menu", "MenuItem"}) then
		writef("X       : %g"&LF, {0} )
		writef("Y       : %g"&LF, {0} )
		writef("CX      : %g"&LF, {0} )
		writef("CY      : %g"&LF, {0} )
	else 
		writef("X       : %g"&LF, {x} )
		writef("Y       : %g"&LF, {y} )
		writef("CX      : %g"&LF, {this[CONTROL_CX]} )
		writef("CY      : %g"&LF, {this[CONTROL_CY]} )
	end if

	if find(this[CONTROL_CLASS], {"Menu", "MenuItem", "Popup", "PopupItem", "Timer"} ) then
		write("Enabled : 1" &LF)
	else
		write("EnableC : 1" &LF)
	end if

	--puts(out_fn, "Checked : 0" & LF)

	if equal( this[CONTROL_CLASS], "TabItem") then
		write("Parent  : " & this[CONTROL_PARENT] &LF)
		ok = created_controls[id][CONTROL_PARENT]
		ok = created_controls[control_by_name(ok)][CONTROL_CLASS]
		write("PClass  : "  & ok &LF)
	elsif sequence(popupinfo) then
		write("Parent  : " &LF)
		write("PClass  : " &LF)	
	else
		ok = find_tabitem_parent(id)
		if ok = 0 then
			ok = ""
		else 
			ok = created_controls[ok][CONTROL_NAME]
		end if
		-- ide needs not to write the parent if the parent is a window
		if equal(ok, "") or
		(equal( created_controls[find_tabitem_parent(id)][CONTROL_CLASS] , "Window")
		and not equal( this[CONTROL_CLASS], "Window")) then
			write("Parent  : " &LF)
			write("PClass  : " &LF)
		else	
			write("Parent  : " & ok &LF)
			ok = created_controls[control_by_name(ok)][CONTROL_CLASS]
			write("PClass  : "  & ok &LF)
		end if
	end if

	write( "Active  : 1" &LF)
	write( "Visible : 1" &LF)	

	write( "Style   : " & this[CONTROL_FLAGS][FLAGS_VALUE] &LF)
	ok = this[CONTROL_FLAGS][FLAGS_STYLE]
	if    ok = NoStyle then
		ok = ""
	elsif ok = SeqStyle then
		ok = "{...}"
	elsif ok = OrAllStyle then
		ok = "or_all({...})"
	end if
	write( "TypStyle: " & ok & LF)

	-- this is not much nice
	ok = this[CONTROL_EXFLAGS][FLAGS_STYLE]
	if    ok = NoStyle then
		ok = this[CONTROL_EXFLAGS][FLAGS_VALUE]
	elsif ok = SeqStyle then
		ok = '{' & this[CONTROL_EXFLAGS][FLAGS_VALUE] & '}'
	elsif ok = OrAllStyle then
		ok = "or_all({" & this[CONTROL_EXFLAGS][FLAGS_VALUE] & "})"
	end if	
	write( "ExStyle : " & ok &LF)
	
	if not xy_default then
		-- yes, this is really reversed
		write( "DefaultXY : 1" &LF)
	end if
	
	if this[CONTROL_SCOPE] = GLOBAL then
		write( "Local : 0" &LF)
	end if

	if find(this[CONTROL_CLASS], {"TabItem", "TabControl"}) then
		writef("TabOrder : %d\n", {tab_controls[tab_parent][2]} )
	end if

	if length(winmain) and
	equal(winmain[WINMAIN_NAME], this[CONTROL_NAME]) then
		if id != 1 then
			convertWarning("WinMain for id != 1 may be problematic"&LF)
		end if
		ok = find( winmain[WINMAIN_TYPE], {"Normal","Maximize","Minimize"})
		if ok then
			ok -= 1
			writef("WinMainOpen : %d"&LF, ok)
		else
			convertWarning("Unknown WinMain type."&LF)
		end if
	end if
	
	write(LF)

	-- now write all events for this control
	for i=1 to length(registered_events) do
		if equal(registered_events[i][EVENT_FORID], this[CONTROL_NAME]) then
			write_event(i)
		end if
	end for
end procedure

----------------------------------------------------------
-- writes the rest of code
----------------------------------------------------------
procedure write_rest()
	integer line, a, b
	sequence temp
	line = 1

	write( "#Proc: (General)"&LF)
	-- a bit of ad :-)
	write( "-------------------------------------------------------"&LF)
	write( "-- IDE Project generated from exw file using" & LF)
	write( "-- " & AppName & " written by Martin Stachon"&LF)
	write( "-------------------------------------------------------"&LF&LF)
	-- write all code that wasn't converted
	for i=1 to length(ignored) do
		for l = line to ignored[i][START_LINE] -1 do
			write(input_file_lines[l])
		end for
		write( input_file_lines[ignored[i][START_LINE]][1..ignored[i][START_CHAR]-1])
		a = ignored[i][END_CHAR]+1
		b = length(input_file_lines[ignored[i][END_LINE]])
		temp = input_file_lines[ignored[i][END_LINE]][a..b]
		if equal(temp, "\n") or equal(temp, "\r\n") then
			-- this alone LF is not needed
		else	
			write(temp)
		end if
		line = ignored[i][END_LINE] + 1
	end for
	for l = line to length(input_file_lines) do
		write( input_file_lines[l])
	end for
	write( LF&"#EndProc"&LF&LF)
end procedure

----------------------------------------------------------
-- writes the whole project
----------------------------------------------------------
global procedure write_project()
	integer win_count
	sequence windows
	sequence sorted_controls

--trace(1)
	-- get count and ids of all created windows
	win_count = 0
	windows = {}
	for i=1 to length(created_controls) do
		if equal(created_controls[i][CONTROL_CLASS], "Window") then
			win_count += 1
			windows &= i
		end if
	end for		

	-- ignore procedures that are assigned to events
	for routine = 1 to length(routines) do
		for event = 1 to length(registered_events) do  	
			if equal(registered_events[event][EVENT_ROUTINE], routines[routine][ROUTINE_NAME]) then
				-- this routine is assigned to an event
				if not length(routines[routine][ROUTINE_END]) then
					convertWarning(sprintf("Routine %s has no end !"&LF, {routines[routine][ROUTINE_NAME]}))
				end if
				ignored &= {routines[routine][ROUTINE_START][1..2] & routines[routine][ROUTINE_END][3..4]}
			end if
		end for
	end for
	ignored = custom_sort(compare_ignored_id, ignored)

	-- assign parent ids for controls, because we will
	-- need it several times
	for i=1 to length(created_controls) do
		if equal(created_controls[i][CONTROL_PARENT], "") then
			created_controls[i][CONTROL_PARENTID] = 0
		else
			created_controls[i][CONTROL_PARENTID] = control_by_name(created_controls[i][CONTROL_PARENT])
		end if
	end for
	
	-- write the header
	write(  "#Version 0.12.0\n")
	writef( "#Counter: %d\n", {length(created_controls)} )
	writef( "#WinCounter: %d\n", {win_count} )

	if win_count = 0 then
		puts(2, "Warning: No window was converted, .prj may cause IDE to crash!\n")
	end if

	-- now all windows and their children
	for win=1 to win_count do
		writef("#Window %d\n", {win})
		write_control(windows[win])
		-- write (General)
		if win = 1 then
			write_rest()
		end if
--		if length(winmain) and
--		equal(winmain[WINMAIN_NAME], created_controls[windows[win]][CONTROL_NAME]) then
--			-- write custom WinMain()
--			write( "#Proc: (Final)" &LF)
--			writef( "WinMain(%s, %s)" & LF, winmain)
--			write( "#EndProc" & LF)
--		end if

		-- now write all controls that have null parent
		-- and aren't window
		-- workaround because ide is window oriented and all
		-- controls must be under a window
		if win=1 then
			for i=1 to length(created_controls) do
				if created_controls[i][CONTROL_PARENTID] = 0 and
				not equal(created_controls[i][CONTROL_CLASS], "Window") then
					write_control(i)
				end if
			end for
		end if
		
		sorted_controls = {}
		-- find all childs of the window
		for c=1 to length(created_controls) do
			if equal(find_top_parent_id(c), windows[win])
			and not find(c, windows) -- we don't put windows here.
			then
				sorted_controls &= c
			end if
		end for
	
		sorted_controls = sort_controls(windows[win] & sorted_controls)
		
		for i=2 to length(sorted_controls) do
			write_control(sorted_controls[i])
		end for
	end for
	close(out_fn)
	
	-- show short summary
	printf(1, "Windows converted : %d\n", {win_count})
	printf(1, "Controls converted : %d\n", {length(created_controls)})
	printf(1, "Events assigned : %d\n", {length(registered_events)})
	
end procedure
